package com.eliasbalasis.tibcopagebus4gwt.client;

import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.juglar.gwt.jsonizer.client.Jsonizer;
import org.juglar.gwt.jsonizer.client.JsonizerParser;

import com.google.gwt.core.client.JavaScriptObject;
import com.google.gwt.user.client.ui.Composite;
import com.google.gwt.user.client.ui.TextArea;

/**
 * This is the main GWT widget which serves as an access point
 * for all <strong>TIBCO PageBus&trade;</strong> functions.
 * </br></br></br>
 * <h1>How can PageBus  help me? </h1>
 * <p>As Ajax  expands the capabilities of Web pages into solutions patterns such as mashups, composite applications, and rich interactive portals, PageBus helps developers drastically simplify development by using publish and subscribe APIs to send events and messages between components rather than get bogged down in point to point integrations between components that lead to unwieldy and less manageable code. In addition, adherence to publish and subscribe architecture helps modularize applications into distinct pieces making team development, unit testing, and source code management easier as well.</p>
 * <p>Publish and subscribe architecture is also the basis for Ajax standards. The first specification from <a href="http://www.openajax.org">The OpenAjax Alliance</a>, an industry group with more than 70 members working to set Ajax interoperability standards, centers on publish and subscribe APIs as its core strategy for facilitating interoperation between Ajax libraries. TIBCO has been an active participant in these processes and has donated the core of the PageBus to the OpenAjax Alliance's implementation of the OpenAjax Hub 1.0. </p>
 * <p><img src="pagebus_figure1.jpg" alt="Ajax Portlets with Reverse Proxy Architecture" width="402" height="256" /></p>
 * This diagram shows three Ajax widgets communicating with each other via the PageBus. Each widget has a file defining its visuals, a file controlling its behavior including the publish and subscribe calls to the PageBus, and a file that gets data from a service on the Web. </p>
 * <h1>How can <b>{@linkplain PageBusAdapter}</b> help me? </h1>
 * Assume now that one or more of the above mentioned Widgets are GWT modules and desire to communicate with the rest of the Ajax widgets on the page.
 * This is a task for {@linkplain PageBusAdapter} which exposes PageBus functions as a GWT widget.
 * Assume that the information to be communicated is represented with the following Java classes </br>
 * <pre class="code" style="margin-left: 2em; padding: 0.5em; border-left: 3px solid #ccc">
 * package com.mycompany.mybeans;
 * ...
 * // a Person bean class.</br>
 * class Person{</br>
 *   ...</br>
 *   public String getName(){...}</br>
 *   public void setName(String name){...}</br>
 *   ...</br>
 *   public int getAge(){...}</br>
 *   public void setAge(int age){...}</br>
 *   ...
 * }</br>
 * 
 * class SubscriberData {</br>
 *   ...</br>
 *   public String getHeader(){...}</br>
 *   public void setHeader(String header){...}</br>
 *   ...</br>
 *   public String getBody(){...}</br>
 *   public void setBody(String header){...}</br>
 *   ...
 * }</br>
 * </pre>
 * as part of the publish and subscribe calls to the PageBus.</br>
 * </br>
 * Not only <code>Person</code> beans must be translated to JavaScript objects before being passed around to PageBus but also JavaScripts objects passed through PageBus by other widgets must be translated to <code>Person</code> beans before the GWT module can use them.</br>
 * PageBusAdapter makes uses marshaller implementations created using the the <a href="http://code.google.com/p/gwt-jsonizer">gwt-jsonizer</a> library to perform the translation.</br>
 * Here is the gwt-jsonizer marshaller implementation for the <code>Person</code> bean.</br>
 * <pre class="code" style="margin-left: 2em; padding: 0.5em; border-left: 3px solid #ccc">
 * interface PersonJsonizer extends Jsonizer{}</br>
 * interface SubscriberDataJsonizer extends Jsonizer{}</br>
 * </pre>
 * It is quite simple don't you think? But you may wonder where is the marshalling/unmarshalling code? It is automatically created by the gwt-jsonizer GWT compiler hook the first time a bean class is accessed.</br>
 * The requirements of the gwt-jsonizer marshaller are</br>
 * <ol>
 * <li>The bean must be compliant with Java Beans specification (getters, setters, default constructors etc.)</li>
 * <li>The bean marshaller definition must be named after the respective bean class name with the <code>Jsonizer</code> suffix</li>
 * <li>
 * Each and every bean property must contain a hint to the marshaller in the form of javadoc annotations. The following annotations are currently supported
 *   <table cellpadding="10px" cellspacing="0" border="1">
 *     <tbody>
 *       <tr><td align="center"><b>Annotation</b></td><td align="center"><b>Effect</b></td></tr>
 *       <tr><td><code>jsonizer.transient</code></td><td>The property in not translatable</td></tr>
 *       <tr><td width="200"><code>jsonizer.propName 'prop'</code></td><td>This property is named 'prop' in JavaScript</td></tr>
 *       <tr><td><code>jsonizer.required</code></td><td>If the property doesn't exist in respetive JavaScript version an exception will be thrown (org.juglar.gwt.jsonizer.client.JsonizerException)</td></tr>
 *       <tr><td><code>gwt.typeArgs <com.mycompany.mybeans.Person></code></td><td>Used for bean properties of Collection type to reference the type of each element in the collection</td></tr>
 *     </tbody>
 *   </table>
 * </li>
 * </ol>
 * </br>
 * Here is how a GWT module using PageBus functions would look like</br>
 * <pre class="code" style="margin-left: 2em; padding: 0.5em; border-left: 3px solid #ccc">
 * package ...;
 * import com.mycompany.myapplication;
 * ...
 * public class MyGWTModule implements EntryPoint {
 * final PageBusAdapter pageBusAdapter = new PageBusAdapter();
 *   ...
 *   RootPanel.add(pageBusAdapter);
 *   SubscriberData subscriberData = new SubscriberData();
 *   subscriberData.setHeader("myHeader");
 *   subscriberData.setBody("myHeader");
 *   // <b>Subscribe to message and associate subsequent receptions with custom subscriber data</b>
 *   pageBusAdapter.PageBusSubscribe("com.mycompany.mymessagesubjects.person", null, null, subscriberData, (Jsonizer)GWT.create(SubscriberDataJsonizer.class));
 *   ...
 *   Person person = new Person();
 *   person.setName("Elias Balasis");
 *   person.setAge(37);
 *   ...
 *   // <b>publish a message with <i>Person</i> bean data</b>
 *   pageBusAdapter.PageBusPublish("com.mycompany.mymessagesubjects.person", person, (Jsonizer)GWT.create(PersonJsonizer.class));
 *   ...
 *   // <b>register listener</b>
 *   pageBusAdapter.addPageBusSubscriptionCallbackListener(
 *     new PageBusListener() {
 *       public void onPageBusSubscriptionCallback(PageBusEvent event) {
 *         // <b>translate JavaScript message contents and subscriber data to their Java equivalents</b>
 *         Person message = (Person)event.getMessage((Jsonizer)GWT.create(PersonJsonizer.class));
 *         SubscriberData subscriberData = (SubscriberData)event.getSubscriberData((Jsonizer)GWT.create(SubscriberDataJsonizer.class));
 *         ...
 *         ...
 *         ...
 *       }
 *     }
 *   );
 *   ...
 *   // <b>unsubscribe from message. future publications of message will not trigger callback</b>
 *   pageBusAdapter.PageBusUnsubscribe("com.mycompany.mymessagesubjects.person");
 * }
 * </pre>
 * </br>
 * </br>
 * <p>
 * <b>Do not forget to include the <code>pagebus.js</code> reference in the HTML file of your GWT module</b>
 * </p>
 * Here is how it should look like
 * </br>
 * <pre class="code" style="margin-left: 2em; padding: 0.5em; border-left: 3px solid #ccc">
 * &lt;html&gt;
 * 
 *   &lt;head&gt;
 *   &lt;!--                                           --&gt;
 *   &lt;!-- Any title is fine                         --&gt;
 *   &lt;!--                                           --&gt;
 *   &lt;title&gt;Wrapper HTML for ...&lt;/title&gt;
 * 
 *   <b>&lt;script src="pagebus.js" language="JavaScript1.2"&gt;&lt;/script&gt;</b>
 *   <b>&lt;script type="text/javascript"></b>
 *     <b>if(window.parent.PageBus) {</b>
 *     <b>window.PageBus = window.parent.PageBus;</b>
 *   <b>}</b>
 *   <b>&lt;/script&gt;</b>
 * 
 *   &lt;!--                                           --&gt;
 *   &lt;!-- The module reference below is the link    --&gt;
 *   &lt;!-- between html and your Web Toolkit module  --&gt;		
 *   &lt;!--                                           --&gt;
 *   &lt;meta name='gwt:module' content='com.mycompany.mygwtapp.MyApp'/&gt;
 * 
 *   &lt;!--                                           --&gt;
 *   &lt;!-- Link CSS file                             --&gt;
 *   &lt;!--                                           --&gt;
 *   &lt;link type="text/css" rel='stylesheet' href='Test.css'/&gt;
 * 
 *   &lt;/head&gt;
 * 
 *   &lt;!--                                           --&gt;
 *   &lt;!-- The body can have arbitrary html, or      --&gt;
 *   &lt;!-- we leave the body empty because we want   --&gt;
 *   &lt;!-- to create a completely dynamic ui         --&gt;
 *   &lt;!--                                           --&gt;
 *   &lt;body&gt;
 * 
 *     &lt;!--                                            --&gt;
 *     &lt;!-- This script is required bootstrap stuff.   --&gt;
 *     &lt;!-- You can put it in the HEAD, but startup    --&gt;
 *     &lt;!-- is slightly faster if you include it here. --&gt;
 *     &lt;!--                                            --&gt;
 *     &lt;script language="javascript" src="gwt.js"&gt;&lt;/script&gt;
 * 
 *     &lt;!-- OPTIONAL: include this if you want history support --&gt;
 *     &lt;iframe id="__gwt_historyFrame" style="width:0;height:0;border:0"&gt;&lt;/iframe&gt;
 * 
 *   &lt;/body&gt;
 * &lt;/html&gt;
 * </pre>
 * <p>
 * <b>Do not forget to include the gwt-jsonizer library reference configuration file (file.gwt.xml) of your GWT module</b>
 * </p>
 * Here is how it should look like
 * </br>
 * <pre class="code" style="margin-left: 2em; padding: 0.5em; border-left: 3px solid #ccc">
 * &lt;module&gt;
 *   &lt;inherits name="com.google.gwt.user.User"/&gt;
 *   <b>&lt;inherits name="com.eliasbalasis.tibcopagebus4gwt.tibcopagebus4gwt"/&gt;</b>
 *   <b>&lt;inherits name="org.juglar.gwt.jsonizer.Jsonizer"/&gt;</b>
 *   ...
 *   ...
 *   ...
 * &lt;/module&gt;
 * </pre>
 */
public final class PageBusAdapter extends Composite implements IPageBusAdapter {
	private TextArea textArea;
	//private static boolean jsLoaded = false;
	private Vector subscriptionCallbackListeners = new Vector();
	private Map subscriptions = new HashMap();
	//private int numLinesLogged;
	private int maxLinesLogged=100;
	private boolean bLogEvents = false;
	private boolean moveCursorOnEvent;

	private final int LOGGER_EVENTTYPE_SUBSCRIBE   = 0;
	private final int LOGGER_EVENTTYPE_PUBLISH     = 1;
	private final int LOGGER_EVENTTYPE_UNSUBSCRIBE = 2;
	private final int LOGGER_EVENTTYPE_CALLBACK    = 3;

	private final String LOGGER_EVENTNAME_SUBSCRIBE   = "SUBSCRIBE";
	private final String LOGGER_EVENTNAME_PUBLISH     = "PUBLISH";
	private final String LOGGER_EVENTNAME_UNSUBSCRIBE = "UNSUBSCRIBE";
	private final String LOGGER_EVENTNAME_CALLBACK    = "CALLBACK";

	/**
	 * default constructor</br>
	 * It is required for integration with Designer tools
	 */
	public PageBusAdapter() {
		/*
		if (jsLoaded==false) {
			Element js = DOM.createElement("script");
			DOM.setElementAttribute(js, "src", "pagebus.js");
			DOM.appendChild(getElement(), js);
			jsLoaded=true;
		}
		*/
		textArea = new TextArea();
		initWidget(textArea);
		textArea.setFocus(true);
		textArea.setSize("100%", "100%");
		textArea.setReadOnly(true);
		setSize("100%", "100%");
	}

	private void firePageBusSubscriptionCallback(String subject, JavaScriptObject message, JavaScriptObject subscriberData) {
		PageBusEvent event = new PageBusEvent(this,subject,message,subscriberData);
		Iterator items = subscriptionCallbackListeners.iterator();
		while(items.hasNext()) {
			PageBusListener item = (PageBusListener)items.next();
			item.onPageBusSubscriptionCallback(event);
			logEvent(LOGGER_EVENTTYPE_CALLBACK, subject, null, event.getMessageAsJsonString(), event.getSubscriberDataAsJsonString(), item.getName());
		}
	}

	private void logEvent(int eventType, String subject, String jsonScope, String jsonMessageData, String jsonSubscriberData, String listenerName) {
		if (bLogEvents==false)
			return;
		String eventMessage = new Date().toString()+" - ";
		int numLines = 0;
		if (eventType==LOGGER_EVENTTYPE_SUBSCRIBE) {
			eventMessage += LOGGER_EVENTNAME_SUBSCRIBE;
		}
		if (eventType==LOGGER_EVENTTYPE_PUBLISH) {
			eventMessage += LOGGER_EVENTNAME_PUBLISH;
		}
		if (eventType==LOGGER_EVENTTYPE_UNSUBSCRIBE) {
			eventMessage += LOGGER_EVENTNAME_UNSUBSCRIBE;
		}
		if (eventType==LOGGER_EVENTTYPE_CALLBACK) {
			eventMessage += LOGGER_EVENTNAME_CALLBACK;
		}
		eventMessage += " - "+subject+"\n";
		++numLines;
		if (eventType==LOGGER_EVENTTYPE_SUBSCRIBE) {
			eventMessage += "\tScope = "+jsonScope+"\n";
			eventMessage += "\tSubscriber Data = "+jsonSubscriberData+"\n";
			++numLines;
		}
		if (eventType==LOGGER_EVENTTYPE_PUBLISH) {
			eventMessage += "\tMessage Data = "+jsonMessageData+"\n";
			++numLines;
		}
		if (eventType==LOGGER_EVENTTYPE_UNSUBSCRIBE) {
		}
		if (eventType==LOGGER_EVENTTYPE_CALLBACK) {
			eventMessage += "\tListener        = "+listenerName+"\n";
			eventMessage += "\tMessage Data    = "+jsonMessageData+"\n";
			eventMessage += "\tSubscriber Data = "+jsonSubscriberData+"\n";
			++numLines;
		}
		String text = getTextArea().getText();
		/*
		String[] lines = text.split("\n");
		int extraLines = 0;
		if (lines.length+numLines>this.maxLinesLogged) {
			extraLines = (lines.length+numLines)-this.maxLinesLogged;
		}
		lines=null;
		while (extraLines>0) {
			text = text.replaceFirst(".*\n", "");
			--extraLines;
		}
		*/
		text += eventMessage;
		getTextArea().setText(text);
		if (isMoveCursorOnEvent()) {
			getTextArea().setCursorPos(getTextArea().getText().length());
		}
	}

	private native void jsPageBusPublish(String subject, JavaScriptObject message)	/*-{
		$wnd.PageBus.publish(subject,message);
	}-*/;

	private native JavaScriptObject jsPageBusSubscribe(String subject, JavaScriptObject scope, JavaScriptObject subscriberData) /*-{
		var me = this;
		var subscription =
			$wnd.PageBus.subscribe(
				subject,
				scope,
				function(callbackSubject, callbackMessage, callbackSubscriberData) {
					me.@com.eliasbalasis.tibcopagebus4gwt.client.PageBusAdapter::firePageBusSubscriptionCallback(Ljava/lang/String;Lcom/google/gwt/core/client/JavaScriptObject;Lcom/google/gwt/core/client/JavaScriptObject;)
					(
					callbackSubject,
					callbackMessage,
					callbackSubscriberData
					);
				},
				subscriberData
			);
		return subscription;
	}-*/;

	private native void jsPageBusUnsubscribe(String subject, JavaScriptObject subscription) /*-{
		$wnd.PageBus.unsubscribe(subscription);
	}-*/;

	private TextArea getTextArea() {
		return textArea;
	}

	/**
	 * Subscribes to PageBus hub for a desired message type
	 * @param subject
	 * <p style="padding-left: 2em;">
	 * The identity of the message type
	 * It can be anything although a package notation is recommended by the PageBus manufacturer</b>
	 * For example <code>com.mycompany.mymessages.notifyChange</code></b>
	 * If <b>null</b> or an empty string is passed subscription will not take place
	 * </p>
	 * @param scope
	 * <p style="padding-left: 2em;">
	 * The message scope. It can be any Java Bean
	 * provided that it is compatible with the marshaller provided in
	 * <code>scopeJsonizer</code> parameter
	 * </p> 
	 * @param scopeJsonizer
	 * <p style="padding-left: 2em;">
	 * The marshaller implementation of the message scope.
	 * It is used to translate the message scope to its JavaScript equivalent
	 * </p>
	 * @param subscriberData
	 * <p style="padding-left: 2em;">
	 * Special data to be included in all future message receptions.
	 * It can be any Java Bean provided that it is compatible
	 * with the marshaller provided in
	 * <code>subscriberDataJsonizer</code> parameter
	 * </p> 
	 * @param subscriberDataJsonizer
	 * <p style="padding-left: 2em;">
	 * The marshaller implementation of the special message data.
	 * It is used to translate the special data
	 * to its JavaScript equivalent
	 * </p>
	 */
	public void PageBusSubscribe(
			String subject,
			Object scope, Jsonizer scopeJsonizer,
			Object subscriberData, Jsonizer subscriberDataJsonizer)
			throws PageBusAdapterException {
		if (subject==null)
			return;
		if (subject.trim().length()==0)
			return;
		PageBusUnsubscribe(subject);
		JavaScriptObject jsScope = null;
		String jsonScope = null;
		if (scope!=null && scopeJsonizer!=null) {
			jsonScope = scopeJsonizer.asString(scope);
			jsScope = JsonizerParser.evaluate(jsonScope);
		}
		JavaScriptObject jsSubscriberData = null;
		String jsonSubscriberData = null;
		if (subscriberData!=null && subscriberDataJsonizer!=null) {
			jsonSubscriberData = subscriberDataJsonizer.asString(subscriberData);
			jsSubscriberData = JsonizerParser.evaluate(jsonSubscriberData);
		}
		JavaScriptObject subscription = null;
		try {
			subscription = jsPageBusSubscribe(subject, jsScope, jsSubscriberData);
		} catch(Exception ex) {
			throw new PageBusAdapterException("Could not subscribe to "+"'"+subject+"'",ex);
		}
		logEvent(LOGGER_EVENTTYPE_SUBSCRIBE, subject, jsonScope, null, jsonSubscriberData, null);
		subscriptions.put(subject, subscription);
	}

	/**
	 * Publishes a new message to the PageBus hub
	 * @param subject
	 * <p style="padding-left: 2em;">
	 * The identity of the message type
	 * It can be anything although a package notation is recommended by the PageBus manufacturer</b>
	 * For example <code>com.mycompany.mymessages.notifyChange</code></b>
	 * If <b>null</b> or an empty string is passed publication will not take place
	 * </p>
	 * @param message
	 * <p style="padding-left: 2em;">
	 * The message data.
	 * It can be any Java Bean provided that it is
	 * compatible with the marshaller
	 * provided in <code>jsonizer</code> parameter
	 * </p> 
	 * @param jsonizer
	 * <p style="padding-left: 2em;">
	 * The marshaller implementation for the message data.
	 * It is used to translate the message data
	 * to its JavaScript equivalent
	 * </p>
	 */
	public void PageBusPublish(
			String subject,
			Object message, Jsonizer jsonizer)
			throws PageBusAdapterException {
		if (subject==null || message==null || jsonizer==null)
			return;
		if (subject.trim().length()==0)
			return;
		String jsonMessage = jsonizer.asString(message);
		logEvent(LOGGER_EVENTTYPE_PUBLISH, subject, null, jsonMessage, null, null);
		try {
			JavaScriptObject jsMessage = JsonizerParser.evaluate(jsonMessage);
			jsPageBusPublish(subject, jsMessage);
		} catch(Exception ex) {
			throw new PageBusAdapterException("Could not publish "+"'"+subject+"'",ex);
		}
	}
	/**
	 * Removes subscription of a desired message type from PageBus hub
	 * @param subject
	 * <p style="padding-left: 2em;">
	 * The identity of the message type
	 * It can be anything although a package notation is recommended by the PageBus manufacturer</b>
	 * For example <code>com.mycompany.mymessages.notifyChange</code></b>
	 * If <b>null</b> or an empty string is passed unsubscribe will not take place
	 * </p>
	 */
	public void PageBusUnsubscribe(String subject) throws PageBusAdapterException {
		if (subject==null)
			return;
		if (subject.trim().length()==0)
			return;
		JavaScriptObject subscription = (JavaScriptObject)subscriptions.get(subject);
		if (subscription==null)
			return;
		try {
			jsPageBusUnsubscribe(subject,subscription);
		} catch(Exception ex) {
			throw new PageBusAdapterException("Could not unsubscribe from "+"'"+subject+"'",ex);
		}
		subscriptions.remove(subject);
		logEvent(LOGGER_EVENTTYPE_UNSUBSCRIBE, subject, null, null, null, null);
	}

	/**
	 * Registers a listener
	 * whose {@linkplain PageBusListener#onPageBusSubscriptionCallback(PageBusEvent)} method
	 * will be called when a message of any subscribed type
	 * is published through the PageBus hub
	 * @param listener
	 * <p style="padding-left: 2em;">
	 * A reference to an instance of
	 * the desired listener implementation class
	 * </p>
	 */
	public void addPageBusSubscriptionCallbackListener(PageBusListener listener) {
		subscriptionCallbackListeners.add(listener);
	}

	/**
	 * Unregisters a listener
	 * @param listener
	 * <p style="padding-left: 2em;">
	 * A reference to an instance of
	 * the desired listener implementation class
	 * </p>
	 */
	public void removePageBusSubscriptionCallbackListener(PageBusListener listener) {
		subscriptionCallbackListeners.remove(listener);
	}

	/**
	 * Gets the maximum number of lines for the displayed events buffer
	 * @return the maxLinesLogged
	 * <p style="padding-left: 2em;">
	 * The maximum number of lines
	 * </p>
	 */
	public int getMaxLinesLogged() {
		return maxLinesLogged;
	}

	/**
	 * Sets the maximum number of lines for the displayed events buffer
	 * @param maxLinesLogged
	 * <p style="padding-left: 2em;">
	 * The maximum number of lines
	 * </p>
	 */
	public void setMaxLinesLogged(int maxLinesLogged) {
		this.maxLinesLogged = maxLinesLogged;
	}

	public void ClearBuffer() {
		this.textArea.setText("");
	}

	/**
	 * Enables or disables the logging of actions in the text area
	 * @param bLogEvents
	 * <b>true</b> enables logging, <b>false</b> disables logging
	 */
	public void EnableLogging(boolean bLogEvents) {
		this.bLogEvents = bLogEvents;
	}

	/**
	 * Reports whether or not actions will be displayed in text area
	 * @return
	 * <b>true</b> when logging is enabled, <b>false</b> otherwise
	 */
	public boolean isLoggingEnabled() {
		return bLogEvents;
	}

	/**
	 * Reports whether cursos in text area moves to last entry when a new event is logged
	 * @return
	 * <b>true</b> moves cursos, <b>false</b> cursos remains at its present location
	 */
	public boolean isMoveCursorOnEvent() {
		return moveCursorOnEvent;
	}

	/**
	 * Enables or disables moving of cursos position in text area to last entry when a new event is logged
	 * @param moveCursorOnEvent
	 * <b>true</b> moves cursos, <b>false</b> cursos remains at its present location
	 */
	public void setMoveCursorOnEvent(boolean moveCursorOnEvent) {
		this.moveCursorOnEvent = moveCursorOnEvent;
	}
}
