package com.eliasbalasis.tibcopagebus4gwt.client;

import java.util.EventObject;
import org.juglar.gwt.jsonizer.client.Jsonizer;
import org.juglar.gwt.jsonizer.client.JsonizerException;
import com.google.gwt.core.client.JavaScriptObject;

/**
 * Instances of this class cannot be created but a new instance is created and passed as parameter to the related event handler ({@linkplain PageBusListener#onPageBusSubscriptionCallback(PageBusEvent)}) <br>
 * Use {@link #getSubject()} to extract the message's unique identity<br>
 * Use {@link #getMessage} method to extract the message data as a java object<br>
 * Use {@link #getSubscriberData} method to extract the data associated upon subscription. ({@link PageBusAdapter#PageBusSubscribe}) <br>
 */
public final class PageBusEvent extends EventObject {
	private String subject;
	private JavaScriptObject javaScriptMessage;
	private JavaScriptObject javaScriptSubscriberData;
	private String  jsonMessage;
	private String jsonSubscriberData;

	/**
	 * Private constructor
	 */
	PageBusEvent(IPageBusAdapter sender, String subject, JavaScriptObject javaScriptMessage, JavaScriptObject javaScriptSubscriberData) {
		super(sender);
		setSubject(subject);
		setJavaScriptMessage(javaScriptMessage);
		setJavaScriptSubscriberData(javaScriptSubscriberData);
	}

	/**
	 * 
	 * @return
	 * A unique identity representing the type of the message assigned. Can be any arbitrary value but package notation format is used in the majority of cases. This is also the practice recommended by TIBCO PageBus, the manufacturer of the underlying JavaScript framework.  <br>
	 * Example <br>
	 * <code>com.mycompany.mypackage.myevent</code> or <code>com.mycompany.mypackage.mysubpackage.myevents.event1</code>
	 */
	public String getSubject() {
		return subject;
	}
	/**
	 * Used internally to assign the message type
	 */
	private void setSubject(String subject) {
		this.subject = subject;
	}
	/**
	 * Used internally to extract the underlying JavaScript object which represents the message data
	 * @return
	 * The message data representation
	 */
	private JavaScriptObject getJavaScriptMessage() {
		return javaScriptMessage;
	}
	/**
	 * Used internally to assign the underlying JavaScript object representing the message data
	 */
	private void setJavaScriptMessage(JavaScriptObject message) {
		this.javaScriptMessage = message;
	}
	/**
	 * Used internally to extract the underlying JavaScript object which represents the subscriber data
	 * @return
	 * The subscriber data representation
	 */
	private JavaScriptObject getJavaScriptSubscriberData() {
		return javaScriptSubscriberData;
	}
	/**
	 * Used internally to assign the underlying JavaScript object representing the subscriber data
	 */
	private void setJavaScriptSubscriberData(JavaScriptObject subscriberData) {
		this.javaScriptSubscriberData = subscriberData;
	}
	/**
	 * A reference to the {@linkplain PageBusAdapter} instance which fired the event
	 * @return
	 * The message sender
	 */
	public IPageBusAdapter getSender() {
		return (IPageBusAdapter)getSource();
	}
	/**
	 * Extracts the java object representation of the message data using a custom marshaller. <br>
	 * The marshaller implementation must match the type of the data associated with the message
	 * otherwise a related exception will be thrown
	 * @param jsonizer
	 * A reference to the marshaller implementation. See {@linkplain PageBusAdapter} for more details
	 * @return
	 * The java object representation of the message data. It must be converted to the associated Java type
	 * @throws PageBusAdapterException
	 * If conversion to Java object fails the thrown exception is expected to contain
	 * the originating exception (org.juglar.gwt.jsonizer.client.JsonizerException)
	 * representing the error in more detail
	 */
	public Object getMessage(Jsonizer jsonizer) throws PageBusAdapterException {
		if (jsonizer==null || getJavaScriptMessage()==null)
			return null;
		Object object=null;
		JavaScriptObject jsMessage = getJavaScriptMessage();
		try {
			object = jsonizer.asJavaObject(jsMessage);
		} catch(JsonizerException ex) {
			throw new PageBusAdapterException("Failed not convert JavaScript object to Java Object",ex);
		}
		try {
			this.jsonMessage = jsonizer.asString(object);
		} catch(JsonizerException ex) {
			throw new PageBusAdapterException("Failed not convert JavaScript object to JSON string",ex);
		}
		return object;
	}
	/**
	 * Extracts the java object representation of the subscriber data using a custom marshaller.<br>
	 * The marshaller implementation must match the type of the data associated with the message
	 * otherwise a related exception will be thrown
	 * @param jsonizer
	 * A reference to the marshaller implementation. See {@linkplain PageBusAdapter} for more details
	 * @return
	 * The java object representation of the message data. It must be converted to the associated Java type
	 * @throws PageBusAdapterException
	 * If conversion to Java object fails the thrown exception is expected to contain
	 * the originating exception (org.juglar.gwt.jsonizer.client.JsonizerException)
	 * representing the error in more detail
	 */
	public Object getSubscriberData(Jsonizer jsonizer) throws PageBusAdapterException {
		if (jsonizer==null || getJavaScriptSubscriberData()==null)
			return null;
		Object object=null;
		JavaScriptObject jsSubscriberData = getJavaScriptSubscriberData();
		try {
			object = jsonizer.asJavaObject(jsSubscriberData);
		} catch(JsonizerException ex) {
			throw new PageBusAdapterException("Failed not convert JavaScript object to Java Object",ex);
		}
		try {
			this.jsonSubscriberData = jsonizer.asString(object);
		} catch(JsonizerException ex) {
			throw new PageBusAdapterException("Failed not convert JavaScript object to JSON string",ex);
		}
		return object;
	}

	String getMessageAsJsonString() {
		return jsonMessage;
	}
	String getSubscriberDataAsJsonString() {
		return jsonSubscriberData;
	}
}
